%%
%
%
%
function rates = eigenfaces(k, no_visuals, length)
    fprintf('Loading data...\n');
    load('ORL_32x32.mat'); % matrix with face images (fea) and labels (gnd)
    load('train_test_orl.mat'); % training and test indices (trainIdx, testIdx)
    fea = double(fea / 255);
    
    if nargin == 1
        no_visuals = false;
    end

    % partition the data into training and test subset
    n_train = size(trainIdx,1);
    n_test = size(testIdx,1);
    train_data = fea(trainIdx,:);
    train_label = gnd(trainIdx,:);
    test_data = fea(testIdx,:);
    test_label = gnd(testIdx,:);
    
    extra_data = zeros(12, 1024);
    extra_data(1,:) = load_vector('./extra/bill.png');
    extra_data(2,:) = load_vector('./extra/corn.png');
    extra_data(3,:) = load_vector('./extra/dollar.png');
    extra_data(4,:) = load_vector('./extra/flower.png');
    extra_data(5,:) = load_vector('./extra/hamburger.png');
    extra_data(6,:) = load_vector('./extra/jobs.png');
    extra_data(7,:) = load_vector('./extra/leaf.png');
    extra_data(8,:) = load_vector('./extra/lime_cat.png');
    extra_data(9,:) = load_vector('./extra/linus.png');
    extra_data(10,:) = load_vector('./extra/mars_face1.png');
    extra_data(11,:) = load_vector('./extra/mars_face2.png');
    extra_data(12,:) = load_vector('./extra/mgb.png');
    extra_data(13,:) = load_vector('./extra/noise.png');
    extra_data(14,:) = load_vector('./extra/orange.png');
    extra_data(15,:) = load_vector('./extra/rms.png');
    
    if ~no_visuals
        display_faces(fea,10,10);
        title('Face data');
        pause;
    end

    mean_face = mean(train_data);   %  >:-(
    
    if ~no_visuals
        imshow(reshape(mean_face, 32, 32) );
        pause;
    end
    
    fprintf('Running PCA...\n');
    [coeffs scores] = princomp(train_data);% find principal components (use princomp function)

    eigen_data = {train_data, test_data, mean_face, coeffs, scores, train_label, test_label};
    %eigen_data = {train_data, extra_data, mean_face, coeffs, scores};
    
    if nargin < 3
        rates = compare_datasets(k, no_visuals, eigen_data)
    end
    
    if nargin > 2
        sample_step = k;
        samples = fix(length/sample_step);
        rates = zeros(samples,1);
        for(i=0:samples)
            rates(i+1) = compare_datasets(i*sample_step+1, no_visuals, eigen_data);
        end
    end
end

function [rate dist] = compare_datasets(k, no_visuals, eigen_data)

    train_data = eigen_data{1};
    test_data = eigen_data{2};
    mean_face = eigen_data{3};
    coeffs = eigen_data{4};
    scores = eigen_data{5};
    have_labels = false;
    train_label = 0;
    test_label = 0;

    if size(eigen_data, 2) > 5
        have_labels = true;
        train_label = eigen_data{6};
        test_label = eigen_data{7};
    end
    
	n_train = size(train_data, 1);
    n_test = size(test_data, 1);
    
    fprintf('Running for k = %i\n', k);    
    
    components = coeffs(:,1:k)';
    show_face_square(no_visuals, components, ceil(sqrt(k)), 'Top principal components');

    
    fprintf('Projecting training data...\n');
    train_data_w = scores(:,1:k);
    train_data_reconstructed = reconstruct_data(n_train, train_data_w, mean_face, components);
    show_face_square(no_visuals, train_data_reconstructed, ceil(sqrt(n_train)), 'Training reconstruction');

    fprintf('Projecting test data...\n');
    test_data_w = calc_weights(n_test, k, mean_face, test_data, components);
    test_data_reconstructed = reconstruct_data(n_test, test_data_w, mean_face, components);
    show_face_square(no_visuals, test_data_reconstructed, ceil(sqrt(n_test)), 'Test reconstruction');
    
    fprintf('Running nearest-neighbor classifier...\n');
    [nn_index dist] = nearest_neighbor(n_test, n_train, test_data_reconstructed, train_data_reconstructed);
    
    %show_face_error(test_data(1:15,:), dist, k);
    
    estimated_label = 0;
    rate = 0;
    if have_labels
        estimated_label = get_labels(n_test, nn_index, train_label);
        rate = sum(estimated_label == test_label)/n_test;
        fprintf('Classification rate: %f\n', rate);
    end
      
    if ~no_visuals
        if have_labels
           show_matches(12, test_data, test_data_reconstructed, train_data, train_data_reconstructed, ...
                             nn_index, dist, estimated_label, test_label);
        else
           show_matches(5, test_data, test_data_reconstructed, train_data, train_data_reconstructed, ...
                             nn_index, dist);
        end
    end
end

%%
% data manipulation

function w = calc_weights(n_test, k, mean_face, test_data, components)
    w = zeros(n_test,k); % low-dim coefficients for test data
    
    for i=1:n_test
        diff_face = test_data(i,:) - mean_face;
        w(i,:) = components * diff_face';
    end
end

function data_reconstructed = reconstruct_data(n_test, w, mean_face, components)  
    data_reconstructed = zeros(n_test, 1024);% high-dimensional reconstructed test faces
    
    for i=1:n_test
        data_reconstructed(i,:) = mean_face + w(i,:) * components;
    end
end

function [nn_ind dist] = nearest_neighbor(n_test, n_train, test_data, train_data)
    nn_ind = zeros(n_test,1);
    min_dist = zeros(n_test,1);

    for i=1:n_test
        train_dist = ones(1024, n_train);
        for j=1:n_train
            train_dist(:,j) = (test_data(i,:) - train_data(j,:)).^2;
        end
        
        train_dist = sum(train_dist);
        [v min_index] = min(train_dist);
        min_dist(i) = v;
        nn_ind(i) = min_index;
    end
    
    if nargout > 1
        dist = min_dist;
    end
end

function estimated_label = get_labels(n_test, nn_index, train_label)
    estimated_label = zeros(n_test,1);
    
    for i=1:n_test
        estimated_label(i) = train_label(nn_index(i));
    end
end


%%
% visualization functions

function show_face_square(no_visuals, face_data, dim, label)
    if ~no_visuals
        display_faces(face_data, dim, dim);
        title (label);
        pause;
    end
end

function show_face_error(face_data, dist, k)
    num_pics = size(face_data,1);
    f = subplot1(1, num_pics);
    for i=1:num_pics
        subplot1(i);
        imshow( reshape(face_data(i,:) , [32 32]) ,[] );
        if i == ceil(num_pics/2)
            title (['k = ' num2str(k)]);
        end
        xlabel( num2str(dist(i), '%.2f') );
    end
    
    saveas(f, ['extra-k_' num2str(k) '.png']);
end

function show_matches(cols, test_data, test_data_reconstructed, train_data, train_data_reconstructed, ...
                             nn_index, dist, estimated_label, real_label)
	current_col = 0;
    
    for img_index = 1:size(test_data, 1)
        
        current_col = current_col + 1;
        
        subplot(4,cols,current_col+cols*0);
        imshow(reshape(test_data(img_index,:),[32 32]),[]);
        if current_col == cols/2; title('Orig. test img.'); end
        
        subplot(4,cols,current_col+cols*1);
        imshow(reshape(test_data_reconstructed(img_index,:),[32 32]),[]);
        if current_col == cols/2; title('Low-dim test img.'); end
        
        subplot(4,cols,current_col+cols*2);
        imshow(reshape(train_data_reconstructed(nn_index(img_index),:),[32 32]),[]);
        xlabel( num2str(dist(img_index), '%.2f') );
        if current_col == cols/2; title('Low-dim nearest neighbor'); end
        
        subplot(4,cols,current_col+cols*3);
        imshow(reshape(train_data(nn_index(img_index),:),[32 32]),[]);
        if current_col == cols/2; title('Orig. nearest neighbor'); end
        
        if nargin > 8
            if estimated_label(img_index)~=real_label(img_index)
                xlabel('incorrect');
            end
        end
        
        if current_col == cols
            pause
            current_col = 0;
            clf;
        end
    end
end
%%

function im = load_vector(name)
    im = double(imread(name));
    im = reshape(im, 1, 1024);
    im = im ./ max(im);
end
